/**
 * Copyright © 2012, 2013 dr. ir. Jeroen M. Valk
 * 
 * This file is part of Badgerfish CPX. Badgerfish CPX is free software:
 * you can redistribute it and/or modify it under the terms of the GNU Lesser
 * General Public License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version. Badgerfish
 * CPX is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along
 * with Badgerfish CPX. If not, see <http://www.gnu.org/licenses/>.
 */

package com.googlecode.aaw.arche;

import org.slf4j.Logger;
import org.slf4j.helpers.FormattingTuple;
import org.slf4j.helpers.MarkerIgnoringBase;
import org.slf4j.helpers.MessageFormatter;
import org.slf4j.spi.LocationAwareLogger;

public abstract class AbstractLogger extends MarkerIgnoringBase implements
		Logger {
	private static final long serialVersionUID = -5176655472848702090L;
	private static final int LOG_LEVEL_TRACE = LocationAwareLogger.TRACE_INT;
	private static final int LOG_LEVEL_DEBUG = LocationAwareLogger.DEBUG_INT;
	private static final int LOG_LEVEL_INFO = LocationAwareLogger.INFO_INT;
	private static final int LOG_LEVEL_WARN = LocationAwareLogger.WARN_INT;
	private static final int LOG_LEVEL_ERROR = LocationAwareLogger.ERROR_INT;
	private static int trace_count = 0;
	private static int debug_count = 0;
	private static int info_count = 0;
	private static int warn_count = 0;
	private static int error_count = 0;

	protected int currentLogLevel = LOG_LEVEL_INFO;

	abstract protected void log_trace(String msg, Throwable t);

	abstract protected void log_debug(String msg, Throwable t);

	abstract protected void log_info(String msg, Throwable t);

	abstract protected void log_warn(String msg, Throwable t);

	abstract protected void log_error(String msg, Throwable t);

	private void log(int level, String message, Throwable t) {
		if (!isLevelEnabled(level)) {
			return;
		}
		switch (level) {
		case LOG_LEVEL_TRACE:
			++trace_count;
			log_trace(message, t);
			break;
		case LOG_LEVEL_DEBUG:
			++debug_count;
			log_debug(message, t);
			break;
		case LOG_LEVEL_INFO:
			++info_count;
			log_info(message, t);
			break;
		case LOG_LEVEL_WARN:
			++warn_count;
			log_warn(message, t);
			break;
		case LOG_LEVEL_ERROR:
			++error_count;
			log_error(message, t);
			break;
		}
	}

	/**
	 * For formatted messages, first substitute arguments and then log.
	 * 
	 * @param level
	 * @param format
	 * @param arg1
	 * @param arg2
	 */
	private void formatAndLog(int level, String format, Object arg1, Object arg2) {
		if (!isLevelEnabled(level)) {
			return;
		}
		FormattingTuple tp = MessageFormatter.format(format, arg1, arg2);
		log(level, tp.getMessage(), tp.getThrowable());
	}

	/**
	 * For formatted messages, first substitute arguments and then log.
	 * 
	 * @param level
	 * @param format
	 * @param arguments
	 *            a list of 3 ore more arguments
	 */
	private void formatAndLog(int level, String format, Object... arguments) {
		if (!isLevelEnabled(level)) {
			return;
		}
		FormattingTuple tp = MessageFormatter.arrayFormat(format, arguments);
		log(level, tp.getMessage(), tp.getThrowable());
	}

	/**
	 * Is the given log level currently enabled?
	 * 
	 * @param logLevel
	 *            is this level enabled?
	 */
	private boolean isLevelEnabled(int logLevel) {
		// log level are numerically ordered so can use simple numeric
		// comparison
		return (logLevel >= currentLogLevel);
	}

	/** Are {@code trace} messages currently enabled? */
	final public boolean isTraceEnabled() {
		return isLevelEnabled(LOG_LEVEL_TRACE);
	}

	final public int getTraceCount() {
		return trace_count;
	}

	/**
	 * A simple implementation which logs messages of level TRACE according to
	 * the format outlined above.
	 */
	final public void trace(String msg) {
		log(LOG_LEVEL_TRACE, msg, null);
	}

	/**
	 * Perform single parameter substitution before logging the message of level
	 * TRACE according to the format outlined above.
	 */
	final public void trace(String format, Object param1) {
		formatAndLog(LOG_LEVEL_TRACE, format, param1, null);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * TRACE according to the format outlined above.
	 */
	final public void trace(String format, Object param1, Object param2) {
		formatAndLog(LOG_LEVEL_TRACE, format, param1, param2);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * TRACE according to the format outlined above.
	 */
	final public void trace(String format, Object... argArray) {
		formatAndLog(LOG_LEVEL_TRACE, format, argArray);
	}

	/** Log a message of level TRACE, including an exception. */
	final public void trace(String msg, Throwable t) {
		log(LOG_LEVEL_TRACE, msg, t);
	}

	/** Are {@code debug} messages currently enabled? */
	final public boolean isDebugEnabled() {
		return isLevelEnabled(LOG_LEVEL_DEBUG);
	}

	final public int getDebugCount() {
		return debug_count;
	}

	/**
	 * A simple implementation which logs messages of level DEBUG according to
	 * the format outlined above.
	 */
	final public void debug(String msg) {
		log(LOG_LEVEL_DEBUG, msg, null);
	}

	/**
	 * Perform single parameter substitution before logging the message of level
	 * DEBUG according to the format outlined above.
	 */
	final public void debug(String format, Object param1) {
		formatAndLog(LOG_LEVEL_DEBUG, format, param1, null);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * DEBUG according to the format outlined above.
	 */
	final public void debug(String format, Object param1, Object param2) {
		formatAndLog(LOG_LEVEL_DEBUG, format, param1, param2);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * DEBUG according to the format outlined above.
	 */
	final public void debug(String format, Object... argArray) {
		formatAndLog(LOG_LEVEL_DEBUG, format, argArray);
	}

	/** Log a message of level DEBUG, including an exception. */
	final public void debug(String msg, Throwable t) {
		log(LOG_LEVEL_DEBUG, msg, t);
	}

	/** Are {@code info} messages currently enabled? */
	final public boolean isInfoEnabled() {
		return isLevelEnabled(LOG_LEVEL_INFO);
	}

	final public int getInfoCount() {
		return info_count;
	}

	/**
	 * A simple implementation which logs messages of level INFO according to
	 * the format outlined above.
	 */
	final public void info(String msg) {
		log(LOG_LEVEL_INFO, msg, null);
	}

	/**
	 * Perform single parameter substitution before logging the message of level
	 * INFO according to the format outlined above.
	 */
	final public void info(String format, Object arg) {
		formatAndLog(LOG_LEVEL_INFO, format, arg, null);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * INFO according to the format outlined above.
	 */
	final public void info(String format, Object arg1, Object arg2) {
		formatAndLog(LOG_LEVEL_INFO, format, arg1, arg2);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * INFO according to the format outlined above.
	 */
	final public void info(String format, Object... argArray) {
		formatAndLog(LOG_LEVEL_INFO, format, argArray);
	}

	/** Log a message of level INFO, including an exception. */
	final public void info(String msg, Throwable t) {
		log(LOG_LEVEL_INFO, msg, t);
	}

	/** Are {@code warn} messages currently enabled? */
	final public boolean isWarnEnabled() {
		return isLevelEnabled(LOG_LEVEL_WARN);
	}

	final public int getWarnCount() {
		return warn_count;
	}

	/**
	 * A simple implementation which always logs messages of level WARN
	 * according to the format outlined above.
	 */
	final public void warn(String msg) {
		log(LOG_LEVEL_WARN, msg, null);
	}

	/**
	 * Perform single parameter substitution before logging the message of level
	 * WARN according to the format outlined above.
	 */
	final public void warn(String format, Object arg) {
		formatAndLog(LOG_LEVEL_WARN, format, arg, null);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * WARN according to the format outlined above.
	 */
	final public void warn(String format, Object arg1, Object arg2) {
		formatAndLog(LOG_LEVEL_WARN, format, arg1, arg2);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * WARN according to the format outlined above.
	 */
	final public void warn(String format, Object... argArray) {
		formatAndLog(LOG_LEVEL_WARN, format, argArray);
	}

	/** Log a message of level WARN, including an exception. */
	final public void warn(String msg, Throwable t) {
		log(LOG_LEVEL_WARN, msg, t);
	}

	/** Are {@code error} messages currently enabled? */
	final public boolean isErrorEnabled() {
		return isLevelEnabled(LOG_LEVEL_ERROR);
	}

	final public int getErrorCount() {
		return error_count;
	}

	/**
	 * A simple implementation which always logs messages of level ERROR
	 * according to the format outlined above.
	 */
	final public void error(String msg) {
		log(LOG_LEVEL_ERROR, msg, null);
	}

	/**
	 * Perform single parameter substitution before logging the message of level
	 * ERROR according to the format outlined above.
	 */
	final public void error(String format, Object arg) {
		formatAndLog(LOG_LEVEL_ERROR, format, arg, null);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * ERROR according to the format outlined above.
	 */
	final public void error(String format, Object arg1, Object arg2) {
		formatAndLog(LOG_LEVEL_ERROR, format, arg1, arg2);
	}

	/**
	 * Perform double parameter substitution before logging the message of level
	 * ERROR according to the format outlined above.
	 */
	final public void error(String format, Object... argArray) {
		formatAndLog(LOG_LEVEL_ERROR, format, argArray);
	}

	/** Log a message of level ERROR, including an exception. */
	final public void error(String msg, Throwable t) {
		log(LOG_LEVEL_ERROR, msg, t);
	}
}
