/**
 * Copyright © 2012, 2013 dr. ir. Jeroen M. Valk
 * 
 * This file is part of Badgerfish CPX. Badgerfish CPX is free software:
 * you can redistribute it and/or modify it under the terms of the GNU Lesser
 * General Public License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version. Badgerfish
 * CPX is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along
 * with Badgerfish CPX. If not, see <http://www.gnu.org/licenses/>.
 */

package com.googlecode.aaw.badgerfish;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;

import org.slf4j.Logger;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;

import com.googlecode.aaw.arche.Arche;

@SuppressWarnings("restriction")
/**
 * This class provides a Badgerfish view on XML documents. That is, a view similar
 * to the way an XML document would look after being transformed into a badgerfish
 * representation. In particular, the exact order between textnodes and elements,
 * and elements with different tagnames is neglected. The focus is on elements by
 * tagname and their order and the concatenated text inside elements.
 * 
 * @author jeroen
 *
 */
final public class Badgerfish {
	static final private Arche arche = Arche.getArche(Badgerfish.class);
	static final private Logger logger = (Logger) arche.getInstance();
	static final private DocumentBuilder db = (DocumentBuilder) arche
			.getInstance();
	static final private XPath xpath = (XPath) arche.getInstance();
	static final private Transformer transformer = (Transformer) arche
			.getInstance();
	static final private BASE64Decoder decoder = new BASE64Decoder();
	static final private BASE64Encoder encoder = new BASE64Encoder();
	private final File base;
	private final Document document;
	private Element element = null;

	public Badgerfish(InputSource source, File base) {
		this.base = base;
		Document document = null;
		try {
			if (source == null) {
				document = db.newDocument();
			} else {
				document = db.parse(source);
			}
		} catch (SAXException e) {
			logger.error(e.getMessage(), e);
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		this.document = document;
	}

	/**
	 * Selects the concatenated text in an xpath-specified element.
	 * 
	 * @param element
	 *            {Element} element inside this document relative to which the
	 *            xpath is evaluated
	 * @param xpath
	 *            {String} an xpath expression as string
	 * @return {String}
	 * 
	 * @throws MojoExecutionException
	 */
	public String selectText(Element element, String xpath) {
		Element text = selectElement(element, xpath);
		if (text == null) {
			return null;
		} else {
		}
		return null;
	}

	/**
	 * Evaluates an xpath expression relative to some element in this Badgerfish
	 * document.
	 * 
	 * @param element
	 * @param xpath
	 * @return
	 * @throws MojoExecutionException
	 */
	public Element selectElement(Element element, String xpath) {
		XPathExpression expr;
		try {
			expr = Badgerfish.xpath.compile(xpath);
		} catch (XPathExpressionException e) {
			logger.error(e.getMessage(), e);
			return null;
		}
		if (element == null) {
			try {
				return (Element) expr.evaluate(document, XPathConstants.NODE);
			} catch (XPathExpressionException e) {
				logger.error(e.getMessage(), e);
			}
		} else {
			try {
				return (Element) expr.evaluate(element, XPathConstants.NODE);
			} catch (XPathExpressionException e) {
				logger.error(e.getMessage(), e);
			}
		}
		return null;
	}

	public void appendChild(Node child) {
		if (element == null) {
			document.appendChild(child);
		} else {
			element.appendChild(child);
		}
	}

	public void startDocument() {
		element = null;
	}

	public void endDocument() {
	}

	public Element startElement(String tagname) {
		Element aux = this.selectElement(element, tagname);
		if (aux == null) {
			aux = document.createElement(tagname);
			this.appendChild(aux);
		}
		element = aux;
		return element;
	}

	public Element endElement() {
		Node node = element.getParentNode();
		if (node.getNodeType() == Node.ELEMENT_NODE) {
			element = (Element) node;
		} else {
			element = null;
		}
		return element;
	}

	public void element(String tagname, String value) {
		Element aux = document.createElement(tagname);
		Text text = document.createTextNode(value);
		this.appendChild(aux);
		aux.appendChild(text);
	}

	public String getConfig() {
		Element config = this.selectElement(null, "badgerfish/config");
		StringWriter writer = new StringWriter();
		ByteArrayOutputStream stream = new ByteArrayOutputStream();
		Source source = new DOMSource(config);
		Result result = new StreamResult(stream);
		try {
			transformer.transform(source, result);
		} catch (TransformerException e) {
			logger.error(e.getMessage(), e);
			return null;
		}
		try {
			writer.close();
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
			return null;
		}
		return encoder.encodeBuffer(stream.toByteArray()).replaceAll("[\r\n]",
				"");
	}

	public String getScript() {
		StringBuilder builder = new StringBuilder();
		builder.append("var CPX=function(classname){return CPX.ROOT.import(classname);};CPX.COUNTER=[0,0,0,0,0,0];CPX.CONF='");
		builder.append(this.getConfig());
		builder.append("';CPX.CLASSES=new Array();CPX.CLOSURE=function(result){function badgerfish(){return result;}return badgerfish;};\n");
		try {
			builder.append(new String(decoder.decodeBuffer(this.selectElement(
					null, "badgerfish/script").getTextContent()), "UTF8"));
		} catch (UnsupportedEncodingException e) {
			logger.error(e.getMessage(), e);
		} catch (DOMException e) {
			logger.error(e.getMessage(), e);
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		builder.append("process.on('exit',function(){process.exit(CPX.COUNTER[0]+CPX.COUNTER[1])});");
		return builder.toString();
	}

	public void element(String tagname, Reader reader) {
		int n;
		StringBuilder builder = new StringBuilder();
		char[] buffer = new char[2048];
		try {
			while ((n = reader.read(buffer)) > 0) {
				builder.append(buffer, 0, n);
			}
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		this.element(tagname,
				encoder.encodeBuffer(builder.toString().getBytes()));

	}

	public File fromFile() {
		File file = new File(base, "/script.js");
		try {
			Writer writer = new FileWriter(file);
			writer.append(this.getScript());
			writer.close();
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		return file;
	}

	public String toString() {
		StringWriter writer = new StringWriter();
		Source source = new DOMSource(document);
		Result result = new StreamResult(writer);
		try {
			transformer.transform(source, result);
		} catch (TransformerException e) {
			logger.error(e.getMessage(), e);
			return null;
		}
		try {
			writer.close();
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
			return null;
		}
		return writer.toString();
	}

	public File toFile() {
		Element temp = (Element) this.selectElement(null,
				"badgerfish/config/outputDirectory");
		if (temp == null) {
			return null;
		} else {
			return new File(temp.getTextContent() + "/classes.xml");
		}
	}
}
